import unittest

import jsonutils as js
from jsonutils.base import JSONBool, JSONDict, JSONInt, JSONList, JSONNull
from jsonutils.exceptions import JSONQueryMultipleValues
from jsonutils.functions.decorators import global_config


class JsonTest(unittest.TestCase):
    def setUp(self):
        js.config.native_types = False
        js.config.query_exceptions = False

    def test_native_types(self):
        test = js.JSONObject(
            {"A": True, "B": None, "D": dict(A=True, B=None), "L": [1, 2]}
        )
        self.assertIsInstance(test.get(A=js.All), JSONBool)
        self.assertIsInstance(test.get(B=js.All), JSONNull)
        self.assertIsInstance(test.get(D=js.All), JSONDict)
        self.assertIsInstance(test.get(L=js.All), JSONList)

        js.config.native_types = True

        self.assertEqual(test.get(A=js.All).__class__, bool)
        self.assertEqual(test.get(B=js.All).__class__, type(None))
        self.assertEqual(type(test.get(D=js.All)), dict)
        self.assertEqual(type(test.get(L=js.All)), list)

        self.assertEqual(test.query(A=js.All).first().__class__, bool)
        self.assertEqual(test.query(A=js.All).last().__class__, bool)
        self.assertEqual(test.query(B=js.All).first().__class__, type(None))
        self.assertEqual(test.query(B=js.All).last().__class__, type(None))

        self.assertEqual(type(test.query(A=True).order_by("A").first()), bool)
        self.assertEqual(type(test.query(B=None).order_by("B").first()), type(None))
        self.assertEqual(type(test.query(D__type=dict).order_by("D").first()), dict)
        self.assertEqual(type(test.query(L__type=list).order_by("L").first()), list)

    def test_global_config(self):
        test = js.JSONObject([dict(A=1, B=2), dict(A=1, B=3)])

        @global_config
        def native_true_exceptions_false(data):
            return data.get(A=1)

        @global_config(native_types=False, query_exceptions=False)
        def native_false_exceptions_false(data):
            return data.get(A=1)

        @global_config(native_types=False, query_exceptions=True)
        def native_false_exceptions_true(data):
            return data.get(A=1)

        @global_config(native_types=True, query_exceptions=False, include_parents=True)
        def native_true_exceptions_false_include_parents_true(data):
            return data.get(A=1)

        self.assertEqual(native_true_exceptions_false(test), 1)
        self.assertNotIsInstance(native_true_exceptions_false(test), JSONInt)

        self.assertEqual(native_false_exceptions_false(test), 1)
        self.assertIsInstance(native_false_exceptions_false(test), JSONInt)

        self.assertRaises(
            JSONQueryMultipleValues, lambda: native_false_exceptions_true(test)
        )

        self.assertEqual(
            native_true_exceptions_false_include_parents_true(test), {"A": 1, "B": 2}
        )
        self.assertNotIsInstance(
            native_true_exceptions_false_include_parents_true(test), JSONDict
        )



import json
import re
import unittest
from datetime import datetime
from pathlib import Path

import jsonutils as js
import pytz
from jsonutils.base import (
    JSONBool,
    JSONCompose,
    JSONDict,
    JSONFloat,
    JSONInt,
    JSONList,
    JSONNode,
    JSONNull,
    JSONObject,
    JSONPath,
    JSONSingleton,
    JSONStr,
    JSONUnknown,
)
from jsonutils.encoders import JSONObjectEncoder
from jsonutils.exceptions import JSONQueryException, JSONQueryMultipleValues
from jsonutils.functions.seekers import empty
from jsonutils.query import All, ExtractYear, QuerySet, SingleQuery, ValuesList

BASE_PATH = Path(js.__file__).parent.resolve()


class JsonTest(unittest.TestCase):
    def setUp(self):

        js.config.native_types = False
        js.config.query_exceptions = True

        self.test1 = JSONObject(
            [
                {"Float": 2.3, "Int": 1, "Str": "string"},
                {"Dict": {"Float": 0.0, "List": [1, 2, 3]}},
            ]
        )
        self.test2 = JSONObject(True)
        self.test3 = JSONObject(
            {"List": (True, False), "Bool": True, "Dict": {"Float": 3.2}}
        )
        self.test4 = JSONObject(
            {"List": [0, 0.1, "str", None], "Dict": {"Bool": True, "None": None}}
        )

        self.test5 = JSONObject(
            [
                {
                    "Float": 1.1,
                    "Dict": {
                        "List": [{"Str": "string1", "List": [None, True, False, 1]}],
                        "Null": None,
                    },
                },
                {
                    "List": [{"Dict": {"Float": "1.2"}, "Bool": True}],
                    "Dict": {"List": [None, {"Str": "string2"}]},
                },
                {"Datetime": "2021-05-01", "Dict": {"Datetime": "2021/06/01"}},
            ]
        )

        self.test6 = JSONObject(
            {
                "position_data": [
                    {"text": "dummy text 1", "pos": [1, 2]},
                    {"text": "dummy text 2", "pos": [3, 2]},
                    {"text": "dummy text 3", "pos": [1, 4]},
                    {"text": "dummy text 4", "pos": [2, 5]},
                    {"text": "dummy text 5", "pos": [4, 1]},
                    {"text": "dummy text 6", "pos": [1, 1, 5]},
                ],
                "timestamp_data": [
                    {"value": 523687, "timestamp": "2021-05-01 08:00:00"},
                    {"value": 523689, "timestamp": "2021-05-01 09:00:00"},
                    {"value": 523787, "timestamp": "2021-05-02 08:30:00"},
                    {"value": 525687, "timestamp": "2021-05-05 18:00:25"},
                ],
                "boolean_data": [
                    {"data": [True, 1]},
                    {"data": [False, None]},
                    {"data": [0, 1]},
                    {"data": (False, True)},
                ],
            }
        )

        self.test7 = JSONObject(
            {
                "candidatos": [
                    {
                        "Nombre completo": "Fhànçaömá",
                        "Pasión": "Música",
                        "€uros": 356.3,
                    },
                    {"Nombre completo": "Üm€", "Pasión": "Comida", "€€": "28.36K €"},
                ]
            }
        )

        self.test8 = JSONObject(
            {
                "data_list": [
                    {"name": "", "id": 0, "list": [], "tuple": (), "none": None},
                    {},
                ],
                "list": [0, 1, 2],
            }
        )

    def test_unknown_types(self):
        test = JSONObject({"data": {"datetime": datetime(2021, 1, 1, 0, 0, 0)}})
        test2 = JSONObject({"data": object})

        self.assertEqual(test.data.datetime, JSONStr("2021-01-01T00:00:00"))
        self.assertTrue(test.check_valid_types())
        self.assertIsInstance(test2.data, JSONUnknown)
        self.assertEqual(
            test2.check_valid_types(),
            (False, [{"data": {"path": ("data",), "type": type}}]),
        )

    def test_types(self):
        """Assert all child types are the correct ones"""

        self.assertIsInstance(self.test1, JSONList)
        self.assertIsInstance(self.test1[0], JSONDict)
        self.assertIsInstance(self.test1[0]["Float"], JSONFloat)
        self.assertIsInstance(self.test1[0]["Int"], JSONInt)
        self.assertIsInstance(self.test1[0]["Str"], JSONStr)
        self.assertIsInstance(self.test1[1], JSONDict)
        self.assertIsInstance(self.test1[1]["Dict"], JSONDict)
        self.assertIsInstance(self.test1[1]["Dict"]["Float"], JSONFloat)
        self.assertIsInstance(self.test1[1]["Dict"]["List"], JSONList)

        self.assertIsInstance(self.test3, JSONDict)
        self.assertIsInstance(self.test3["List"], JSONList)
        self.assertIsInstance(self.test3["List"][0], JSONBool)
        self.assertIsInstance(self.test3["List"][1], JSONBool)
        self.assertIsInstance(self.test3["Bool"], JSONBool)
        self.assertIsInstance(self.test3["Dict"], JSONDict)
        self.assertIsInstance(self.test3["Dict"]["Float"], JSONFloat)

        self.assertIsInstance(self.test4, JSONDict)
        self.assertIsInstance(self.test4["List"], JSONList)
        self.assertIsInstance(self.test4["List"][0], JSONInt)
        self.assertIsInstance(self.test4["List"][1], JSONFloat)
        self.assertIsInstance(self.test4["List"][2], JSONStr)
        self.assertIsInstance(self.test4["List"][3], JSONNull)
        self.assertIsInstance(self.test4["Dict"], JSONDict)
        self.assertIsInstance(self.test4["Dict"]["Bool"], JSONBool)
        self.assertIsInstance(self.test4["Dict"]["None"], JSONNull)

    def test_append_method(self):

        test1 = self.test1.copy()
        test2 = self.test5.copy()

        test1.append({"fake": True})
        test2.append(test1._1.Dict.List, serialize_nodes=True)

        self.assertEqual(test2[-1], test1._1.Dict.List)
        self.assertNotEqual(test2[-1]._0.jsonpath, test1._1.Dict.List._0.jsonpath)
        self.assertEqual(test2[-1]._0.jsonpath, "3/0")

        test2.append(test1._1.Dict.List, serialize_nodes=False)
        # if not serializing nodes, the new appended jsonpath will not match the path within test2 jsonobject
        self.assertEqual(test2[-1], test1._1.Dict.List)
        self.assertEqual(test2[-1]._0.jsonpath, test1._1.Dict.List._0.jsonpath)

        self.assertIsInstance(test1, JSONList)
        self.assertIsInstance(test1[0], JSONDict)
        self.assertIsInstance(test1[0]["Float"], JSONFloat)
        self.assertIsInstance(test1[0]["Int"], JSONInt)
        self.assertIsInstance(test1[0]["Str"], JSONStr)
        self.assertIsInstance(test1[1], JSONDict)
        self.assertIsInstance(test1[1]["Dict"], JSONDict)
        self.assertIsInstance(test1[1]["Dict"]["Float"], JSONFloat)
        self.assertIsInstance(test1[1]["Dict"]["List"], JSONList)
        self.assertIsInstance(test1[2], JSONDict)
        self.assertIsInstance(test1[2]["fake"], JSONBool)

        self.assertEqual(test1[0].parent, test1)
        self.assertEqual(test1[1].parent, test1)
        self.assertEqual(test1[2].parent, test1)

        self.assertEqual(test1[0]._index, 0)
        self.assertEqual(test1[1]._index, 1)
        self.assertEqual(test1[2]._index, 2)

        self.assertEqual(test1[0]["Float"]._key, "Float")
        self.assertEqual(test1[2]["fake"]._key, "fake")
        self.assertEqual(test1[2]["fake"].jsonpath, JSONPath("2/fake/"))

        self.assertTrue(test1.query(fake=True).exists())
        self.assertEqual(test1.query_key(".*ak.*").keys().first(), "fake")

    def test_list_set_items(self):

        test1 = self.test1.copy()

        test1[1] = ["A", {"B": None}]

        self.assertIsInstance(test1, JSONList)
        self.assertIsInstance(test1[0], JSONDict)
        self.assertIsInstance(test1[0]["Float"], JSONFloat)
        self.assertIsInstance(test1[0]["Int"], JSONInt)
        self.assertIsInstance(test1[0]["Str"], JSONStr)
        self.assertIsInstance(test1[1], JSONList)
        self.assertIsInstance(test1[1][0], JSONStr)
        self.assertIsInstance(test1[1][1], JSONDict)
        self.assertIsInstance(test1[1][1]["B"], JSONNull)

    def test_dict_set_items(self):

        test3 = self.test3.copy()

        test3["new_key"] = {"key": [None, "None"]}

        self.assertIsInstance(test3, JSONDict)
        self.assertIsInstance(test3["List"], JSONList)
        self.assertIsInstance(test3["List"][0], JSONBool)
        self.assertIsInstance(test3["List"][1], JSONBool)
        self.assertIsInstance(test3["Bool"], JSONBool)
        self.assertIsInstance(test3["Dict"], JSONDict)
        self.assertIsInstance(test3["Dict"]["Float"], JSONFloat)
        self.assertIsInstance(test3["new_key"], JSONDict)
        self.assertIsInstance(test3["new_key"]["key"], JSONList)
        self.assertIsInstance(test3["new_key"]["key"][0], JSONNull)
        self.assertIsInstance(test3["new_key"]["key"][1], JSONStr)

        self.assertEqual(test3["List"].parent, test3)
        self.assertEqual(test3["Bool"].parent, test3)
        self.assertEqual(test3["new_key"].parent, test3)
        self.assertEqual(test3["new_key"]["key"].parent, test3["new_key"])
        self.assertEqual(test3["new_key"]["key"][0].parent, test3["new_key"]["key"])

        self.assertEqual(
            test3["new_key"]["key"][0].jsonpath, JSONPath("new_key/key/0/")
        )

    def test_keys(self):
        """
        Check that all nodes coming directly from dictionary keys have their corresponding key associated with them.
        """

        # Nodes coming from a list do not have an associated key, but an "index" attribute.
        self.assertEqual(self.test1[0]._key, None)
        self.assertEqual(self.test1[1]._key, None)
        self.assertEqual(self.test1[0]._index, 0)
        self.assertEqual(self.test1[1]._index, 1)
        # --------------------------------------------------------------------------------
        self.assertEqual(self.test1[0]["Float"]._key, "Float")
        self.assertEqual(self.test1[0]["Int"]._key, "Int")
        self.assertEqual(self.test1[0]["Str"]._key, "Str")
        self.assertEqual(self.test1[1]["Dict"]._key, "Dict")
        self.assertEqual(self.test1[1]["Dict"]["Float"]._key, "Float")
        self.assertEqual(self.test1[1]["Dict"]["List"]._key, "List")

    def test_parents(self):
        """Check every child object has the right parent object"""

        # the parent of the root node must be None
        self.assertEqual(self.test1.parent, None)

        # the parent of both two list elements must be the list itself
        self.assertEqual(
            self.test1[0].parent,
            JSONList(
                [
                    {"Float": 2.3, "Int": 1, "Str": "string"},
                    {"Dict": {"Float": 0.0, "List": [1, 2, 3]}},
                ]
            ),
        )
        self.assertEqual(
            self.test1[1].parent,
            JSONList(
                [
                    {"Float": 2.3, "Int": 1, "Str": "string"},
                    {"Dict": {"Float": 0.0, "List": [1, 2, 3]}},
                ]
            ),
        )
        # -----------------------------------------
        self.assertEqual(
            self.test1[0]["Float"].parent,
            JSONDict({"Float": 2.3, "Int": 1, "Str": "string"}),
        )
        self.assertEqual(
            self.test1[1]["Dict"].parent,
            JSONDict({"Dict": {"Float": 0.0, "List": [1, 2, 3]}}),
        )
        self.assertEqual(
            self.test1[1]["Dict"]["Float"].parent,
            JSONDict({"Float": 0.0, "List": [1, 2, 3]}),
        )
        self.assertEqual(
            self.test1[1]["Dict"]["List"].parent,
            JSONDict({"Float": 0.0, "List": [1, 2, 3]}),
        )
        self.assertEqual(
            self.test1[1]["Dict"]["List"][0].parent,
            JSONList([1, 2, 3]),
        )
        self.assertEqual(
            self.test1[1]["Dict"]["List"][1].parent,
            JSONList([1, 2, 3]),
        )

    def test_roots_in_queryset(self):
        self.assertEqual(self.test6.query(data__1=True)._root, self.test6)
        self.assertEqual(
            self.test6.position_data.query(pos__1="44")._root, self.test6.position_data
        )

    def test_root(self):
        """
        Verify that any child nodes within the json structure share the same unique root
        """
        self.assertEqual(self.test6.position_data._0.root, self.test6)
        self.assertEqual(self.test6.position_data._1.root, self.test6)
        self.assertEqual(self.test6.root, self.test6)

    def test_paths(self):

        self.assertEqual(
            self.test1.query(List__contains=2).first().jsonpath,
            JSONPath("1/Dict/List/"),
        )
        self.assertEqual(
            self.test1.query(List__contains=2).first().jsonpath.expr,
            '[1]["Dict"]["List"]',
        )

    def test_json_serializable(self):
        """Assert that the JSONObject is serializable"""

        self.assertEqual(
            json.dumps(self.test1).replace('"', "'"), self.test1.__repr__()
        )
        self.assertEqual(
            json.dumps(self.test2, cls=JSONObjectEncoder).replace('"', "'"),
            self.test2.__repr__().lower(),
        )
        self.assertEqual(
            json.dumps(self.test3, cls=JSONObjectEncoder).replace('"', "'"),
            self.test3.__repr__().replace("True", "true").replace("False", "false"),
        )

        self.assertEqual(
            json.dumps(self.test4, cls=JSONObjectEncoder),
            '{"List": [0, 0.1, "str", null], "Dict": {"Bool": true, "None": null}}',
        )

    def test_get(self):
        test = self.test6
        test2 = self.test1

        self.assertEqual(
            test.get(timestamp="2021-05-01 08:00:00"),
            test.query(timestamp="2021-05-01 08:00:00").first(),
        )
        self.assertEqual(test.get(fakeee=1, default_=5), 5)

        self.assertRaises(JSONQueryMultipleValues, lambda: test.get(pos__0=1))

        self.assertEqual(test2.get(Float=2.3), test2.query(Float="2.3").first())
        self.assertRaises(JSONQueryMultipleValues, lambda: test2.get(Float__gt="-1"))
        self.assertRaises(JSONQueryException, lambda: test2.get(Float__gt="-1"))

        self.assertRaisesRegex(
            JSONQueryException,
            "The query has not returned any result",
            lambda: test.get(timestamp="1895-01-01"),
        )

    def test_queries(self):

        test = JSONObject.open(BASE_PATH / "tests/balance-sheet-example-test.json")

        # check isnull query
        self.assertEqual(JSONObject(dict(A=True)).query(A__isnull=True), [])
        self.assertEqual(JSONObject(dict(A=True)).query(A__isnull=False), [True])
        self.assertEqual(JSONObject(dict(A=False)).query(A__isnull=True), [])
        self.assertEqual(JSONObject(dict(A=False)).query(A__isnull=False), [False])
        # ------------------
        self.assertEqual(self.test3.query(fakke="true").first(), JSONNull(None))
        self.assertEqual(self.test3.query(fakke="true").last(), JSONNull(None))
        self.assertEqual(self.test3.query(Bool="true"), [JSONBool(True)])
        self.assertEqual(self.test3.query(Bool__type=str), [])
        self.assertEqual(self.test3.query(Bool__type="str"), [])
        self.assertEqual(self.test3.query(Bool__type=bool), [JSONBool(True)])
        self.assertEqual(self.test3.query(Bool__type="bool"), [JSONBool(True)])
        self.assertEqual(self.test4.query(None__type=None), [JSONNull(None)])
        self.assertEqual(self.test4.query(None__type="None"), [JSONNull(None)])
        self.assertEqual(self.test5.query(Str__type="str"), ["string1", "string2"])
        self.assertEqual(self.test5.query(Str__type="None"), [])
        self.assertEqual(self.test5.query(Str__length=7), ["string1", "string2"])
        self.assertEqual(self.test5.query(Str__length=0), [])
        self.assertEqual(self.test5.query(Str__startswith="st"), ["string1", "string2"])
        self.assertEqual(self.test5.query(Str__startswith="fake"), [])
        self.assertEqual(self.test5.query(Str__endswith=1), ["string1"])
        self.assertEqual(self.test5.query(Str__endswith=2), ["string2"])
        self.assertEqual(self.test5.query(Str__endswith=3), [])
        self.assertEqual(self.test5.query(Str__endswith=(1, 2)), ["string1", "string2"])
        self.assertEqual(
            self.test5.query(Str__endswith=(1, "2")), ["string1", "string2"]
        )
        self.assertEqual(
            self.test5.query(Str__endswith=["1", 2]), ["string1", "string2"]
        )
        self.assertEqual(
            self.test5.query(Str__endswith=["1", "2"]), ["string1", "string2"]
        )
        self.assertEqual(
            self.test5.query(Str__type="singleton"), ["string1", "string2"]
        )
        self.assertEqual(
            self.test5.query(Datetime__type=datetime), self.test5.query(Datetime=All)
        )
        self.assertRaises(JSONQueryException, lambda: self.test5.query(Str__length="1"))
        self.assertEqual(
            self.test3.query(List__contains=True), [[JSONBool(True), JSONBool(False)]]
        )
        self.assertEqual(
            self.test1.query(Dict__contains=["Float", "List"]).last(),
            {"Float": 0.0, "List": [1, 2, 3]},
        )
        self.assertEqual(
            self.test4.query(Dict__in=("Bool", "None", "Other")),
            [{"Bool": True, "None": None}],
        )
        self.assertEqual(
            self.test4.query(Dict__in=("Bool", "Other")),
            [],
        )
        self.assertEqual(
            self.test4.query(List__in=("0", "0.1", "str", None)),
            [[0, 0.1, "str", None]],
        )
        self.assertEqual(
            self.test5.query(Bool__contains=True),
            [True],
        )
        self.assertEqual(self.test5.query(Float=1.2), QuerySet([JSONStr(1.2)]))
        self.assertEqual(
            self.test5.query(Float__gt=1), QuerySet([JSONFloat(1.1), JSONStr(1.2)])
        )
        self.assertEqual(self.test5.query(Float__gt=2), QuerySet())
        self.assertEqual(self.test5.query(Str__exact="string2"), QuerySet(["string2"]))
        self.assertEqual(
            self.test5.query(List=[None, True, False, 1], include_parent_=True),
            [JSONDict({"Str": "string1", "List": [None, True, False, 1]})],
        )
        self.assertEqual(
            self.test5.query(Str__exact="string2", include_parent_=True),
            [{"Str": "string2"}],
        )
        self.assertEqual(
            self.test5.query(List__0__contains=("Dict", "Bool")),
            [[{"Dict": {"Float": "1.2"}, "Bool": True}]],
        )
        self.assertEqual(
            self.test5.query(List__0__contains="Str"),
            [[{"Str": "string1", "List": [None, JSONBool(True), False, 1]}]],
        )
        self.assertEqual(
            self.test5.query(Datetime__gt="2021-05-01"), QuerySet(["2021/06/01"])
        )
        self.assertEqual(
            self.test5.query(Datetime__contains="2021-05"),
            QuerySet([JSONStr("2021-05-01")]),
        )
        self.assertEqual(
            self.test5.query(List__parent__contains="Str"),
            [
                [JSONNull(None), JSONBool(True), JSONBool(False), 1],
            ],
        )
        self.assertEqual(
            self.test6.query(
                timestamp__gt="2021-05-01 08:30:00",
                timestamp__lte="2021-05-02 08:30:00",
                include_parent_=True,
            ),
            [
                {"value": 523689, "timestamp": "2021-05-01 09:00:00"},
                {"value": 523787, "timestamp": "2021-05-02 08:30:00"},
            ],
        )
        self.assertEqual(self.test6.query(pos__0__gte=2), [[3, 2], [2, 5], [4, 1]])
        self.assertEqual(
            self.test6.query(pos__2__gt=2, include_parent_=True),
            [{"text": "dummy text 6", "pos": [1, 1, 5]}],
        )
        self.assertEqual(self.test6.query(pos__5__gte=2), [])
        self.assertEqual(
            self.test6.query(pos__0__gte=2, pos__1__lt=5), [[3, 2], [4, 1]]
        )
        self.assertEqual(
            self.test6.query(pos__1__in=(1, 4), include_parent_=True),
            [
                {"text": "dummy text 3", "pos": [1, 4]},
                {"text": "dummy text 5", "pos": [4, 1]},
                {"text": "dummy text 6", "pos": [1, 1, 5]},
            ],
        )
        self.assertEqual(self.test8.query(list__isnull=True), QuerySet([[]]))
        self.assertEqual(self.test8.query(tuple__isnull=True), QuerySet([[]]))
        self.assertEqual(self.test8.query(none__isnull=True), QuerySet([None]))
        self.assertEqual(self.test8.query(id__isnull=True), QuerySet())
        self.assertEqual(self.test6.query(data__1=True), QuerySet([[False, True]]))

        self.assertListEqual(
            test.query(
                filing_date__isnull=False, filing_date__notpath="all_company_filings"
            )
            .order_by("-filing_date")
            .values(date="filing_date", form_type="form_type", url="form_url"),
            [
                {"date": "2017-09-27", "form_type": "1-A/A", "url": None},
                {"date": "2017-09-06", "form_type": "1-A", "url": None},
            ],
        )

    def test_multiqueries(self):
        test = JSONObject(
            [
                {
                    "data": {
                        "data": [
                            1,
                            2,
                            {"data": 1, "date": "2021/08/01  12:00:00Z"},
                            {"data": {"data": 1, "date": "2021-08-02"}},
                        ],
                        "date": "2021/08/01  13:00:00.003Z",
                    },
                    "date": datetime(2021, 9, 8),
                }
            ]
        )

        self.assertEqual(
            self.test6.query(
                text__regex=r"(?:2|5)", pos__0__gte=2, include_parent_=True
            ),
            [
                {"text": "dummy text 2", "pos": [3, 2]},
                {"text": "dummy text 5", "pos": [4, 1]},
            ],
        )

        self.assertEqual(
            test.query(data=1, date__gte=datetime(2021, 8, 1), include_parent_=True),
            [
                {"data": 1, "date": "2021/08/01  12:00:00Z"},
                {"data": 1, "date": "2021-08-02"},
            ],
        )
        self.assertEqual(
            test.query(data__c_data__contains=1),
            [
                {
                    "data": [
                        1,
                        2,
                        {"data": 1, "date": "2021/08/01  12:00:00Z"},
                        {"data": {"data": 1, "date": "2021-08-02"}},
                    ],
                    "date": "2021/08/01  13:00:00.003Z",
                },
                {"data": 1, "date": "2021-08-02"},
            ],
        )
        self.assertEqual(
            test.annotate(date=None).query(
                data__c_data__contains=1, date__isnull=True, include_parent_=True
            ),
            [{"data": {"data": 1, "date": "2021-08-02"}, "date": None}],
        )

    def test_queries_traversing(self):

        test = self.test6

        self.assertListEqual(
            test.query(
                data__parent__parent__0__c_data__1__exact="1", include_parent_=True
            ),
            [
                {"data": [True, 1]},
                {"data": [False, None]},
                {"data": [0, 1]},
                {"data": [False, True]},
            ],
        )
        self.assertEqual(
            test.query(
                timestamp__lte="2021-05-01T23:59:00",
                timestamp__parent__c_value__contains=89,
                include_parent_=True,
            ),
            [{"value": 523689, "timestamp": "2021-05-01 09:00:00"}],
        )

    def test_single_query_exact(self):

        test_dict = JSONObject({"A": {"A": 1, "B": True}})
        test_list = JSONObject({"A": [1, "2", True, "false"]})
        test_str = JSONObject({"A": "lorep ipsum", "B": {"B1": [1, 2]}})
        test_bool = JSONObject({"A": True})

        self.assertTrue(
            SingleQuery("A", {"A": 1, "B": True})._check_against_node(test_dict.A)
        )
        self.assertFalse(SingleQuery("A", ["A", "B"])._check_against_node(test_dict.A))

        self.assertTrue(SingleQuery("A", "lorep ipsum")._check_against_node(test_str.A))
        self.assertFalse(SingleQuery("A", "lorepipsum")._check_against_node(test_str.A))
        self.assertTrue(
            SingleQuery("A__exact", "lorep ipsum")._check_against_node(test_str.A)
        )
        self.assertFalse(
            SingleQuery("A__exact", "lorepipsum")._check_against_node(test_str.A)
        )

        self.assertTrue(
            SingleQuery("A", [1, "2", True, "false"])._check_against_node(test_list.A)
        )
        self.assertTrue(
            SingleQuery("A", [1, 2, True, False])._check_against_node(test_list.A)
        )
        self.assertTrue(
            SingleQuery(
                "B1__parent__parent__c_A__exact", "lorep ipsum"
            )._check_against_node(test_str.B.B1)
        )

    def test_single_query_gt(self):
        test_float = JSONObject({"A": 1.3})
        test_list = JSONObject({"A": [2, 3], "B": []})

        self.assertTrue(SingleQuery("A__gt", 1)._check_against_node(test_float.A))
        self.assertTrue(SingleQuery("A__gt", [1, 2])._check_against_node(test_list.A))
        self.assertTrue(
            SingleQuery("A__gt", [1, 2, 5])._check_against_node(test_list.A)
        )
        self.assertTrue(SingleQuery("A__gt", [1])._check_against_node(test_list.A))
        self.assertFalse(SingleQuery("A__gt", 1)._check_against_node(test_list.A))
        self.assertFalse(SingleQuery("A__gt", [])._check_against_node(test_list.A))
        self.assertFalse(SingleQuery("A__gt", type)._check_against_node(test_list.A))
        self.assertFalse(SingleQuery("A__gt", [3, 2])._check_against_node(test_list.A))
        self.assertFalse(
            SingleQuery("A__gt", [1, 3, 5])._check_against_node(test_list.A)
        )
        self.assertFalse(
            SingleQuery("B__gt", [1, 3, 5])._check_against_node(test_list.B)
        )
        self.assertFalse(SingleQuery("B__gt", [])._check_against_node(test_list.B))

    def test_single_queries(self):
        test = JSONObject(
            [{"A": [1, 2], "B": {"A": "123"}}, {"date": "2021-05-04T09:08:00"}]
        )

        q1 = SingleQuery("A__contains", 1)
        q2 = SingleQuery("A__in", (0, 2, 1, 3))
        q3 = SingleQuery("A__in", (0, 2))
        q4 = SingleQuery("A__contains", 3)
        q5 = SingleQuery("date", datetime(2021, 5, 4, 9, 8, 0, tzinfo=pytz.utc))
        q6 = SingleQuery("B__c_A", 123)

        self.assertTrue(q1._check_against_node(test._0.A))
        self.assertTrue(q2._check_against_node(test._0.A))
        self.assertFalse(q3._check_against_node(test._0.A))
        self.assertTrue(q4._check_against_node(test._0.B.A))
        self.assertTrue(q5._check_against_node(test._1.date))
        self.assertTrue(q6._check_against_node(test._0.B))

    def test_update_in_nodes(self):
        test = JSONObject(
            {
                "data": [
                    {"name": "Dan", "id": [1, 2]},
                    {
                        "name": "Don",
                        "id": [0, 3],
                        "comments": [{"id": 1, "value": "Great"}],
                    },
                ],
                "timestamp": "2021-01-01 10:00:00Z",
                "valuation": {"pre": 10, "post": 9},
            }
        )
        # update timestamp
        pre_path = test.get(timestamp__year=2021).jsonpath

        self.assertEqual(
            test.get(timestamp__year=2021).update(lambda x: "2022" + x[4:]), True
        )
        self.assertEqual(test.get(timestamp__year=2022), "2022-01-01 10:00:00Z")
        self.assertEqual(test.get(timestamp__year=2021, throw_exceptions_=False), None)
        self.assertEqual(pre_path, test.timestamp.jsonpath)

        # failing updating timestamp

        self.assertEqual(test.timestamp.update(lambda x: x ** 2), False)
        self.assertEqual(test.get(timestamp__year=2022), "2022-01-01 10:00:00Z")

        # update id
        self.assertEqual(test.get(id__0=1).update(lambda x: x + ["fake"]), True)

    def test_update_in_queries(self):

        test = JSONObject(
            {
                "data": {
                    "cik": "0005852352",
                    "fileNumber": "052-065820",
                    "team": [
                        {"name": "David", "bio": "An enthusiast", "date": "2021-05-08"},
                        {"name": "Alex", "bio": "A musician", "date": "2020-01-03"},
                    ],
                },
                "comments": [
                    {
                        "date": "2021-01-01T08:00:00",
                        "text": "Ouuu yeahhh",
                        "responses": 0,
                    },
                    {
                        "date": "2021-01-02T09:30:00",
                        "text": "That sounds quite good",
                        "responses": 22,
                    },
                ],
            }
        )
        test2 = JSONObject(dict(A=1, B=2))
        test6 = self.test6.copy()

        self.assertEqual(test2.query(A=All).update_ifnonnull(None), (0, 1))
        self.assertDictEqual(test2, dict(A=1, B=2))
        self.assertEqual(
            test2.query(A=All).update_ifnonnull(lambda x: None if x == 1 else 3), (0, 1)
        )
        self.assertDictEqual(test2, dict(A=1, B=2))
        self.assertEqual(
            test2.query(A=All).update_ifnonnull(lambda x: 0 if x + 1 == 2 else 3),
            (1, 0),
        )
        self.assertDictEqual(test2, dict(A=0, B=2))

        self.assertEqual(test6.query(data__0=True).update("OK"), (1, 0))
        self.assertEqual(
            test.data.team.query(name__contains="e").update("Veronica"), (1, 0)
        )

        self.assertEqual(test.data.team.query(name__contains="e"), ["Veronica"])
        self.assertEqual(test6.query(data__0=True), [])
        self.assertEqual(test6.query(data="OK"), ["OK"])
        self.assertEqual(
            test6.query(pos__2__gt=3, text__contains=6).update(lambda x: x[:2]), (1, 0)
        )
        self.assertEqual(test6.position_data._5.pos, [1, 1])
        self.assertEqual(test6.position_data.query(pos__2=All), [])
        self.assertEqual(test6.position_data.query(pos=(1, 1)), [[1, 1]])

    def test_all_queries(self):
        test = JSONObject({"A": {"A": 1, "B": 2}, "B": ["A", {"A": 2}]})

        self.assertEqual(test.query(A=All), [{"A": 1, "B": 2}, "$1", 2])
        self.assertEqual(test.query(B=All), [2, ["A", {"A": 2}]])

    def test_setters(self):

        test = JSONObject({"data": [{"A": 1, "B": 2}]})
        test2 = JSONObject({"key": "mykey", "index": 1, "nested": {"index": "2"}})
        test3 = JSONObject([1, 2, {"A": [3, 4]}])

        self.assertEqual(
            test.query(data__0__contains="B"), QuerySet([[{"A": 1, "B": 2}]])
        )

        test.data._0.B = 555

        self.assertFalse(test.query(B=2).exists())
        self.assertEqual(test.query(B=555), [555])

        self.assertEqual(test2.key, "mykey")
        self.assertEqual(test2.index, "1")
        self.assertEqual(test2.nested.index, 2)

        test2.key = 111
        test2.index = 222
        test2.nested.index = 333

        self.assertDictEqual(
            test2, JSONObject({"key": 111, "index": 222, "nested": {"index": 333}})
        )
        self.assertEqual(test2.query(key=All), [111])
        self.assertListEqual(test2.query(index=All), [333, 222])
        self.assertEqual(test2.query(index=333).first().jsonpath, "nested/index")

        self.assertTrue(test3.query(A__contains=4).exists())

        test3._0 = "first"
        test3._1 = "second"
        test3._2.A._1 = "last"

        self.assertListEqual(test3, JSONObject(["first", "second", {"A": [3, "last"]}]))
        self.assertFalse(test3.query(A__contains=4).exists())
        self.assertTrue(test3.query(A__contains="last").exists())

    def test_annotations(self):

        test = self.test1.copy()
        test2 = self.test1.copy()
        test3 = JSONObject([1, 2, {"A": [1, 2, {"B": 3}], "B": 4}])

        self.assertEqual(
            test.annotate(a1=1, a2=2),
            JSONObject(
                [
                    {"Float": 2.3, "Int": 1, "Str": "string", "a1": 1, "a2": 2},
                    {
                        "Dict": {"Float": 0.0, "List": [1, 2, 3], "a1": 1, "a2": 2},
                        "a1": 1,
                        "a2": 2,
                    },
                ]
            ),
        )

        self.assertEqual(test._1.Dict.a2.jsonpath, "1/Dict/a2")
        self.assertEqual(test.query(a1=1).count(), 3)
        self.assertNotEqual(test, self.test1)

        self.assertEqual(
            test2.annotate(a1={"status": "OK"}).query(a1__contains="status"),
            [{"status": "OK"}, {"status": "OK"}, {"status": "OK"}],
        )

        self.assertEqual(
            test3.annotate(C=1, D=2),
            JSONObject(
                [1, 2, {"A": [1, 2, {"B": 3, "C": 1, "D": 2}], "B": 4, "C": 1, "D": 2}]
            ),
        )

        # now we remove the annotations and check if recovers original object

        test._remove_annotations()

        self.assertEqual(test, self.test1)
        self.assertFalse(test.query(a1=All).exists())

    def test_pop(self):

        test = JSONObject({"data": [{"name": "Dan", "age": 30}]})

        self.assertSetEqual(set(test.data._0._child_objects.values()), {"Dan", 30})
        self.assertEqual(test.data._0.name, "Dan")

        # remove a child

        test.data._0.pop("name")

        self.assertSetEqual(set(test.data._0._child_objects.values()), {30})
        self.assertFalse(test.query(name=All).exists())
        self.assertIsNone(test.data._0.name)

    def test_multiparent(self):
        test = JSONObject(
            {
                "root": {
                    "root_list": [0, {"child": {"A": 1}}, {"child2": {"A": 1}}],
                    "root_dict": {"child": {"A": 1}, "key": [0, "key"]},
                }
            }
        )

        self.assertEqual(test.query(A__parents__c_key__1__contains="ey"), ["1"])
        self.assertEqual(test.query(A__parents__c_key__1__contains="ey", A=1), ["1"])
        self.assertFalse(
            test.query(A__parents__c_key__1__contains="ey", A__gt=1).exists()
        )
        self.assertEqual(test.query(A__parents__c_key__1__contains="ey").count(), 1)
        self.assertEqual(
            test.query(A__parents__c_key__1__contains="ey").first().jsonpath,
            "root/root_dict/child/A",
        )
        self.assertEqual(
            test.get(A__parents__0=0, A__parents__index=1, throw_exceptions_=True), 1
        )
        self.assertEqual(
            test.get(
                A__parents__0=0, A__parents__index=1, throw_exceptions_=True
            ).jsonpath,
            "root/root_list/1/child/A",
        )
        self.assertIsNone(
            test.get(
                A__parents__0=0,
                A__parents__index=0,
                throw_exceptions_=False,
                native_types_=True,
            )
        )
        self.assertRaisesRegex(
            JSONQueryException,
            "Lookup parents can only be included once",
            lambda: test.get(A__parents__parents__0=0),
        )

    def test_values(self):

        js.config.query_exceptions = False

        test = JSONObject(
            {
                "id": 1234,
                "company": {
                    "name": "NAME",
                    "filing_set": [
                        {
                            "date": "2021-01-01",
                            "data": {"field1": 1, "field2": 2},
                            "type": "C",
                        },
                        {
                            "date": "2022-01-01",
                            "data": {"field1": 3, "field2": 4},
                            "type": "A",
                        },
                    ],
                },
                "date": "2022-02-02",
                "filing": {"field1": 5},
            }
        )
        test2 = JSONObject(
            {
                "data": {
                    "id": "0008523621",
                    "amount": 55856212.25,
                    "success": False,
                    "team": [
                        {
                            "name": "David G. García",
                            "age": 36,
                            "hobbies": ["music", "reading"],
                        },
                        {
                            "name": "Alexander D. Diego",
                            "age": 31,
                            "hobbies": ["cooking", "music", "sports", "yoga"],
                        },
                        {"name": "David A. Márquez", "age": 28, "hobbies": None},
                        {
                            "name": "Peter C. Jackson",
                            "age": 31,
                            "hobbies": ["sports", "reading", "tennis", "smoke"],
                            "staff": True,
                        },
                    ],
                },
                "updated_at": "2021-05-01T09:30:00",
                "comments": {
                    "data": "structured",
                    "amount": 2,
                    "comments": [
                        0,
                        {
                            "name": "Peter K. Márquez",
                            "text": "Great!",
                            "valuation": 10,
                            "timestamp": "2020-03-01",
                        },
                        1,
                        {
                            "name": "Jacob H. Stack",
                            "text": "Bad!",
                            "valuation": 1,
                            "timestamp": "2021-03-05",
                            "staff": True,
                        },
                    ],
                },
            }
        )

        self.assertEqual(
            test.get(field1=1, field1__parents__c_type="C").values("date"),
            {"date": "2021-01-01"},
        )
        self.assertEqual(
            test.get(field1=1, field1__parents__c_type="C").values("date", "name"),
            {"date": "2021-01-01", "name": "NAME"},
        )
        self.assertEqual(
            test.get(field1=1, field1__parents__c_type="C").values(
                datetime="date", new_name="name"
            ),
            {"datetime": "2021-01-01", "new_name": "NAME"},
        )
        self.assertEqual(
            test.get(field1=1, field1__parents__c_type="C").values(
                "date", new_name="name"
            ),
            {"date": "2021-01-01", "new_name": "NAME"},
        )
        self.assertEqual(
            test.get(
                field1__parents__c_date__year=2022, field1__notpath="filing"
            ).values("date", "type", "id"),
            {"date": "2021-01-01", "type": "C", "id": 1234},
        )
        self.assertEqual(
            test.get(field1__parents__c_date__year=2022, field1__notpath="filing")
            .values("date", "type", "id")
            .id,
            1234,
        )

        self.assertEqual(
            test2.get(staff=True).values("timestamp", "age"),
            {"timestamp": None, "age": 31},
        )
        self.assertEqual(
            test2.get(fake=True).values("timestamp", "age"),
            {"timestamp": None, "age": None},
        )
        self.assertIsNone(
            test2.get(fake=True).values("timestamp", "age").timestamp,
        )

    def test_distinct(self):

        self.assertEqual(
            self.test6.query(timestamp__year=2021)
            .order_by("-timestamp")
            .distinct(lambda x: ExtractYear(x).year),
            ["2021-05-05 18:00:25"],
        )

    def test_delete(self):
        test = JSONObject(
            [
                {
                    "data": [{"name": "Dan", "age": 30}, {"name": "Hel", "age": 31}],
                    "date": datetime(2021, 1, 1),
                },
                {"data": {"name": "Fel", "age": "27"}},
            ]
        )

        self.assertEqual(test.query(date__year=2021).delete(), 1)
        self.assertEqual(test.query(date__year=2021), [])
        self.assertEqual(test.query(data=All).delete(), 2)
        self.assertEqual(test.query(data=All), [])
        self.assertEqual(test, [{}, {}])

    def test_query_keys(self):

        test = JSONObject({"name": {"NAME": "Dan", "name": True}})

        self.assertListEqual(
            test.query_key(js.I(".*na.*")).distinct(lambda x: x._key).keys(),
            ValuesList(["name", "NAME"]),
        )

        self.assertListEqual(
            test.query_key(".*ame", exact=All), [{"NAME": "Dan", "name": True}, True]
        )
        self.assertListEqual(
            test.query_key(re.compile(".*AmE", re.I), exact=All),
            [{"NAME": "Dan", "name": True}, "Dan", True],
        )
        self.assertListEqual(
            test.query_key(re.compile(".*AmE", re.I), type__=bool),
            QuerySet([True]),
        )
        self.assertListEqual(
            test.query_key(
                js.I(".*AmE"),
                type__=bool,
                parent={"NAME": "Dan", "name": True},
            ),
            QuerySet([True]),
        )
        self.assertListEqual(
            test.query_key(
                re.compile(".*AmE", re.I),
                type__=bool,
                parent__key="name",
            ),
            QuerySet([True]),
        )
        self.assertListEqual(
            test.query_key(
                re.compile(".*AmE", re.I),
                type__=bool,
                path="fake",
            ),
            QuerySet([]),
        )

    def test_get_key(self):
        test = self.test5

        self.assertEqual(test.get_key(js.I("flo.*"), exact="1.1"), 1.1)
        self.assertEqual(
            test.get_key("Lis.*", __1=True), JSONList([None, True, False, 1])
        )
        self.assertEqual(
            test.get_key(js.I("flo.*"), gte="1", throw_exceptions_=False), 1.1
        )
        self.assertEqual(test.get_key("FAKEKEY", default_=5), 5)
        self.assertRaises(JSONQueryMultipleValues, lambda: test.get_key("Flo.*", gte=1))
        self.assertRaises(JSONQueryException, lambda: test.get_key("FakeKey"))

    def test_traverse_json(self):
        # TODO what if a key has an explicit "
        test = JSONObject([{"A": 1, "B": {"B1": 2, "B2": [3, 4]}}])

        self.assertListEqual(
            test.traverse_json().values("path", flat=True),
            [
                [
                    0,
                ],
                [0, "A"],
                [0, "B"],
                [0, "B", "B1"],
                [0, "B", "B2"],
                [0, "B", "B2", 0],
                [0, "B", "B2", 1],
            ],
        )

    def test_json_paths(self):
        test = JSONObject({"A": {"B": [{"C": {"D": 1}}]}})

        self.assertEqual(test.A.B._0.C.D.jsonpath.keys, ("A", "B", 0, "C", "D"))
        self.assertEqual(test.A.B._0.C.jsonpath.expr, '["A"]["B"][0]["C"]')

        test.set_path(("A", "B", 0, "D"), 10)
        self.assertTupleEqual(
            test.query_key("*", exact=10).first().jsonpath.keys, ("A", "B", 0, "D")
        )
        test.set_path(("B",), 20)
        self.assertTupleEqual(
            test.query_key("*", exact=20).first().jsonpath.keys, ("B",)
        )

        self.assertRaises(TypeError, lambda: test.set_path(("A", "B", "A"), 30))

    def test_path_exists(self):
        test = JSONObject({"A": {"B": [{"C": {"D": 1}}]}})

        self.assertTrue(test.path_exists(("A",)))
        self.assertTrue(test.path_exists(("A", "B")))
        self.assertTrue(test.path_exists(("A", "B", 0)))
        self.assertTrue(test.path_exists(("A", "B", 0, "C")))
        self.assertTrue(test.path_exists(("A", "B", 0, "C", "D")))

        self.assertFalse(test.path_exists(("a",)))
        self.assertFalse(test.path_exists(("A", 0)))
        self.assertFalse(test.path_exists(("A", "B", 1)))
        self.assertFalse(test.path_exists(("A", "B", 0, "c")))
        self.assertFalse(test.path_exists(("A", "B", 0, "C", "d")))

    def test_eval_path(self):
        test = JSONObject({"A": {"B": [{"C": 1}]}})

        self.assertTrue(test.eval_path(("A",)), {"B": [{"C": 1}]})
        self.assertTrue(test.eval_path("A", {"B": [{"C": 1}]}))
        self.assertTrue(test.eval_path(("A", "B")), [{"C": 1}])
        self.assertTrue(test.eval_path(("A", "B", 0)), {"C": 1})
        self.assertTrue(test.eval_path(("A", "B", 0, "C")), 1)
        self.assertEqual(test.eval_path(("A", "B", 0, "C")).jsonpath, "A/B/0/C")
        self.assertRaises(
            AttributeError,  # jsonpath attribute should not exist when returning native types
            lambda: test.eval_path(("A", "B", 0, "C"), native_types_=True).jsonpath,
        )

        # ---- FAIL PATHS ----
        self.assertRaises(KeyError, lambda: test.eval_path(("B",)))
        self.assertRaises(KeyError, lambda: test.eval_path(("A", "C")))
        self.assertRaises(IndexError, lambda: test.eval_path(("A", "B", 1)))
        self.assertEqual(
            test.eval_path(("A", "A"), fail_silently=True), empty
        )  # not raises an error if fail_silently is True
        self.assertEqual(
            test.eval_path(("A", "A"), fail_silently=True, value_on_exception=99), 99
        )
        self.assertEqual(
            test.eval_path(("A", "A"), fail_silently=True, value_on_exception=0), 0
        )
        self.assertEqual(
            test.eval_path((0, 0), fail_silently=True, value_on_exception=""), ""
        )
        self.assertEqual(
            test.eval_path((0, 0), fail_silently=True, value_on_exception=None), None
        )

    def test_filter(self):
        test_queryset = QuerySet(
            [JSONObject(dict(A=dict(B=1))), dict(A=dict(B=2)), dict(C=1)],
            list_of_root_nodes=True,
        )
        self.assertEqual(
            test_queryset.filter(B=1), QuerySet([JSONObject(dict(A=dict(B=1)))])
        )
        self.assertEqual(
            test_queryset.filter(B__gte=1),
            QuerySet([JSONObject(dict(A=dict(B=1))), JSONObject(dict(A=dict(B=2)))]),
        )
        self.assertEqual(test_queryset.filter(C=All), [dict(C=1)])

    def test_apply(self):
        test = self.test1
        test2 = JSONObject({"A": 1, "B": {"C": 2, "D": 3}})

        self.assertEqual(test.query(Float=All).apply(lambda x: x + 1), [3.3, 1])
        self.assertEqual(test2.query(D=All).apply(lambda x: x + x.parent.C), [5])

    def test_rename_keys(self):
        test = JSONObject(dict(A=1, B=dict(C=2, D=3)))

        self.assertDictEqual(test.rename_keys(A="newA"), dict(newA=1, B=dict(C=2, D=3)))
        self.assertDictEqual(test.rename_keys(B="newB"), dict(A=1, newB=dict(C=2, D=3)))
        self.assertDictEqual(
            test.rename_keys({"A": "newA"}), dict(newA=1, B=dict(C=2, D=3))
        )
        self.assertDictEqual(
            test.rename_keys({"fake": "newA"}), dict(A=1, B=dict(C=2, D=3))
        )

        test.rename_keys(A="newAinplace", inplace=True)

        self.assertDictEqual(test, dict(newAinplace=1, B=dict(C=2, D=3)))
        self.assertEqual(test.newAinplace.jsonpath.keys, ("newAinplace",))



"""
This module contains the base objects of the JSON structure
"""
import json
import sys
from datetime import date, datetime
from pathlib import Path
from uuid import uuid4

import requests
from bs4 import BeautifulSoup

import jsonutils.config as config
from jsonutils.cache import memoized_method
from jsonutils.encoders import JSONObjectEncoder
from jsonutils.exceptions import (
    JSONDecodeException,
    JSONNotFoundException,
    JSONQueryException,
    JSONQueryMultipleValues,
)
from jsonutils.functions.decorators import (
    catch_exceptions,
    dummy,
    return_value_on_exception,
)
from jsonutils.functions.parsers import (
    _parse_html_table,
    _parse_query,
    _parse_query_key,
    _to_django_model,
    parse_bool,
    parse_datetime,
    parse_float,
    parse_timestamp,
    url_validator,
)
from jsonutils.functions.seekers import (
    _eval_object,
    _json_from_path,
    _relative_to,
    _set_object,
    empty,
)
from jsonutils.query import All, KeyQuerySet, ParentList, QuerySet
from jsonutils.utils.dict import (
    UUIDdict,
    ValuesDict,
    _rename_keys,
    _rename_keys_inplace,
)
from jsonutils.utils.retry import retry_function


# ---- external modules ----
def DjangoQuerySet():

    try:
        return sys.modules["django"].db.models.QuerySet
    except Exception:
        return type(None)


def PandasDataFrame():

    try:
        return sys.modules["pandas"].DataFrame
    except Exception:
        return type(None)


# ----------------------


class JSONPath:
    """
    Object representing a JSON path for a given JSON object.
    Don't instanciate it directly.

    Attributes
    ----------
        data: a pretty string representation of the path. Ex: 'data/0/name'.
        expr: a python string representation of the path. Ex: '["data"][0]["name"]'.
        keys: a tuple representation of the path. Ex: ("data", 0, "name").
    """

    @staticmethod
    def _cast_to_int(s):
        if s.isdigit():
            return int(s)
        else:
            return s

    def __init__(self, path=()):
        if isinstance(
            path, str
        ):  # if path is given as an string like 'A/0/B', cast it to a tuple like ("A", 0, "B")
            path = tuple(self._cast_to_int(i) for i in path.split("/") if i != "")
        if not isinstance(path, tuple):
            raise TypeError(
                f"Argument 'path' must be a tuple or str instance, not {type(path)}"
            )
        self._keys = path  # tuple of dict keys

    @property
    def data(self):
        return "/".join(map(str, self._keys))

    @property
    def expr(self):
        output = ""
        for k in self._keys:
            if isinstance(k, str):
                output += f'["{k}"]'
            elif isinstance(k, int):
                output += f"[{k}]"
        return output

    @property
    def keys(self):
        """Returns a tuple with the object's path keys"""
        return self._keys

    def relative_to(self, node):
        """
        Calculate jsonpath relative to a parent node's jsonpath.
        Examples of relative paths:
            * self -> ("A", 0, "B", "C")
                                            ==> ("B", "C")
            * node -> ("A", 0)
        Returns a new JSONPath instance
        """

        if not isinstance(node, JSONNode):
            raise TypeError(
                f"child argument must be a JSONNode instance, not {type(node)}"
            )

        self_path = self._keys
        other_path = node.jsonpath._keys

        result = _relative_to(self_path, other_path)
        jsonpath = JSONPath(result)
        return jsonpath

    def _update(self, **kwargs):
        if (key := kwargs.get("key")) is not None:
            self._keys = (key,) + self._keys
        elif (index := kwargs.get("index")) is not None:
            self._keys = (index,) + self._keys

    def __eq__(self, other):
        if isinstance(other, (tuple, list)):
            return self._keys == tuple(other)
        if isinstance(other, str):
            return self.data == other
        if isinstance(other, self.__class__):
            return self._keys == other._keys
        return NotImplemented

    def __repr__(self):
        return self.data


class JSONObject:
    """
    This class acts as a switcher. It will return the corresponding object instance for a given data.
    This class does not contain any instances of.
    """

    _RESERVED_ATTRIBUTES = (
        "_key",
        "_index",
        "parent",
        "_id",
        "_child_objects",
        "_is_annotation",
    )

    def __new__(cls, data=None, raise_exception=False, serialize_nodes=False):
        """
        Params:
        ------
            data: the json data that is going to be parsed.
            raise_exception: if True, then an exception will be raised if data type is not known. Otherwise, a JSONUnknown
                instance will be created for such a data.
            serialize_nodes: if True and data represents a node instance, then take it as a raw data, discarding
                all its old node attributes (like jsonpath, parent, etc).
        """
        if not isinstance(raise_exception, bool):
            raise TypeError(
                f"raise_exception argument must be a boolean, not {type(raise_exception)}"
            )
        if isinstance(data, JSONNode):
            if serialize_nodes:
                return cls(data._data)
            else:
                return data
        elif isinstance(data, type(None)):
            return JSONNull(data)
        elif isinstance(data, dict):
            return JSONDict(data)
        elif isinstance(data, bool):
            return JSONBool(data)
        elif isinstance(data, str):
            return JSONStr(data)
        elif isinstance(data, (date, datetime)):
            return JSONStr(data.isoformat())
        elif isinstance(data, float):
            return JSONFloat(data)
        elif isinstance(data, int):
            return JSONInt(data)
        # data from external libraries
        elif isinstance(data, (list, tuple, DjangoQuerySet())):
            return JSONList(data)
        elif isinstance(data, PandasDataFrame()):
            return JSONDict(json.loads(data.to_json()))
        else:
            if not raise_exception:
                return JSONUnknown(data)
            else:
                raise JSONDecodeException(f"Wrong data's format: {type(data)}")

    @classmethod
    def open(cls, file, raise_exception=True, **kwargs):
        """
        Open an external JSON file.
        If a valid url string is passed, then it will try to make a get/post request to such a target and decode a json file
        """
        # decide whether to use requests.get or requests.post by checking kwargs
        if kwargs.get("json") or kwargs.get("data"):
            FUNCTION = requests.post
        else:
            FUNCTION = requests.get
        file = str(file)
        if url_validator(file):
            req = retry_function(
                FUNCTION, file, raise_exception=raise_exception, **kwargs
            )
            try:
                data = req.json()
            except Exception as e:
                raise JSONDecodeException(
                    f"Selected URL has no valid json file. Details: {e}"
                )
            else:
                return cls(data)
        with open(file) as f:
            data = json.load(f)
        return cls(data)

    @classmethod
    def loads(cls, string, **kwargs):
        """
        This is a wrapper for json.loads. It takes a json string as argument and returns a JSONNode instance
        """

        try:
            data = json.loads(string, **kwargs)
        except Exception as e:
            raise JSONDecodeException(
                f"Error when parsing the json string. Error message: {e}"
            )
        else:
            return cls(data)

    @staticmethod
    def read_html_table(
        data,
        parser="lxml",
        raise_exception=True,
        attrs={},
        recursive=True,
        parse_links=False,
        link_prefix=None,
        **kwargs,
    ):
        """
        Loads an html table from an url or string as node instance.

        Params
        ------
            data: an string representing the raw data. If an url is passed,
                  then it will try to make a request to the web and locate a valid html table.
                  If, on the other hand, the text string represents an html structure,
                  it will act on it without making any request.
            parser: parse library to be used by `BeautifulSoup`.
            raise_exception: if True, then any errors derived from the table lookup will be thrown as response.
                             If False, it will return None on errors.
            attrs: a dictionary to be passed as an argument to the `find` function of `BeautifulSoup` object.
                   It is useful, for example, if you want to specify a particular attribute that must be present in the HTML tags of the table.
            recursive: it will be passed as an argument to the `find` function of `BeautifulSoup` object.
                       Specifies whether the table lookup should be recursive or not.
            parse_links: if True, then instead of taking the text inside a tag, it will retrieve the first href link, if it exists.
            link_prefix: when `parse_links` is selected, we can prepend an string to href links.
            kwargs: extra arguments to be passed to the request, such as headers, proxies, etc.
        """
        # TODO needs a test
        if not isinstance(data, str):
            raise TypeError(
                f"Argument data must be an 'str' instance, not {type(data)}"
            )

        if url_validator(data):
            try:
                req = retry_function(requests.get, data, **kwargs)
            except Exception:
                if raise_exception:
                    raise
                else:
                    return

            soup = BeautifulSoup(req.content, parser)
            table = soup.find("table", attrs, recursive)

            if not table:
                if raise_exception:
                    raise JSONNotFoundException(
                        "A table matching the required parameters has not been found on the selected website."
                    )
                else:
                    return

            try:
                result = _parse_html_table(
                    table,
                    parse_links,
                    link_prefix,
                )
            except Exception:
                if raise_exception:
                    raise
                else:
                    result = None

            return result

        else:  # data is already an html string
            soup = BeautifulSoup(data, parser)
            table = soup.find("table", attrs, recursive)

            if not table:
                if raise_exception:
                    raise JSONNotFoundException(
                        "A table matching the required parameters has not been found on the selected website."
                    )
                else:
                    return

            try:
                result = _parse_html_table(
                    table,
                    parse_links,
                    link_prefix,
                )
            except Exception:
                if raise_exception:
                    raise
                else:
                    result = None

            return result

    @classmethod
    def from_path(cls, iterable):
        """
        Build a JSONObject from a list of leaf node paths
        """

        obj = _json_from_path(iterable)
        return cls(obj)


class JSONNode:
    """
    This is the base class for all JSON objects (compose or singleton).

    Attributes:
    -----------
        _child_objects:
        _key: last dict parent key where the object comes from
        _index: las list parent index where the object comes from
        parent: last parent object where this object comes from
        _id: unique uuid4 hex identifier of object
    """

    __odir__ = object.__dir__  # rename old __dir__ method to __odir__
    __osetattr__ = object.__setattr__

    def __init__(self, *args, **kwargs):

        self._key = None
        self._index = None
        self.parent = None
        self._id = uuid4().hex

    def _set_new_uuid(self):

        self._id = uuid4().hex
        return self._id

    def json_encode(self, **kwargs):
        return json.dumps(self, cls=JSONObjectEncoder, **kwargs)

    @property
    def is_leaf(self):
        """Check if this node is a leaf node (no childs)"""
        try:
            child_number = self._child_objects.__len__()
        except AttributeError:
            return True
        return not bool(child_number)

    @property
    def json_decode(self):
        return json.loads(json.dumps(self, cls=JSONObjectEncoder))

    @property
    def jsonpath(self):
        parent = self
        path = JSONPath()
        while parent is not None:
            path._update(key=parent._key, index=parent._index)
            parent = parent.parent
        return path

    @property
    def parent_list(self):
        pl = ParentList()

        parent = self
        last = parent
        while parent is not None:
            last = parent
            parent = parent.parent
            pl.append(last)

        return pl[1:]

    @property
    def root(self):
        """Get root object from current node object"""

        parent = self
        last = parent
        while parent is not None:
            last = parent
            parent = parent.parent
        return last

    def update(self, new_obj):
        """
        Update this node within JSONObject from which it is derived
        """

        is_callable = callable(new_obj)

        path = self.jsonpath
        root = self.root

        if is_callable:
            try:
                root.set_path(path, new_obj(root.eval_path(path)))
            except Exception:
                return False
            else:
                return True
        else:
            root.set_path(path, new_obj)
        return True

    def values(self, *keys, search_upwards=True, flat=False, **kwargs):
        if flat is True and len(keys) > 1:
            raise ValueError(
                "If flat argument is selected, you can only set one key value"
            )
        output_dict = ValuesDict({k: None for k in keys})

        if kwargs:
            output_dict.update({k: None for k in kwargs})

        for key in keys:
            obj = self
            if "__" in key:  # traverse by childs in this case
                child_key = key.split("__")[-1]
                if not child_key:
                    raise ValueError("Wrong syntax within query values request")
                output_dict[key] = obj.parent.get(
                    **{child_key: All}, native_types_=True, throw_exceptions_=False
                )
                continue
            while True:
                if isinstance(obj, JSONDict):
                    if key in obj:
                        output_dict[key] = obj[key]._data
                        break
                if (
                    search_upwards
                ):  # TODO search_upwards can be a number (number of recursive upwards lookings)
                    obj = obj.parent
                    if obj is None:
                        break
                else:
                    break

        if kwargs:
            for k, v in kwargs.items():
                obj = self
                if "__" in v:  # traverse by childs in this case
                    child_key = v.split("__")[-1]
                    if not child_key:
                        raise ValueError("Wrong syntax within query values request")
                    output_dict[k] = obj.parent.get(
                        **{child_key: All}, native_types_=True, throw_exceptions_=False
                    )
                    continue
                while True:
                    if isinstance(obj, JSONDict):
                        if v in obj:
                            output_dict[k] = obj[v]._data
                            break
                    if search_upwards:
                        obj = obj.parent
                        if obj is None:
                            break
                    else:
                        break

        return output_dict if not flat else list(output_dict.values())[0]

    @catch_exceptions
    def apply(self, func, fail_silently=False):
        res = func(self)
        if isinstance(res, JSONNode):
            res = res._data
        return res

    def __str__(self):
        return self.json_encode(indent=4, ensure_ascii=False)

    # ---- ACTION METHODS ----
    def gt_action(self, other):
        from jsonutils.functions.actions import _gt

        return _gt(self, other)

    def gte_action(self, other):
        from jsonutils.functions.actions import _gte

        return _gte(self, other)

    def lt_action(self, other):
        from jsonutils.functions.actions import _lt

        return _lt(self, other)

    def lte_action(self, other):
        from jsonutils.functions.actions import _lte

        return _lte(self, other)

    def exact_action(self, other):
        from jsonutils.functions.actions import _exact

        return _exact(self, other)

    def contains_action(self, other):
        from jsonutils.functions.actions import _contains

        return _contains(self, other)

    def icontains_action(self, other):
        from jsonutils.functions.actions import _icontains

        return _icontains(self, other)

    def in_action(self, other):
        from jsonutils.functions.actions import _in

        return _in(self, other)

    def regex_action(self, other):
        from jsonutils.functions.actions import _regex

        return _regex(self, other)

    def fullregex_action(self, other):
        from jsonutils.functions.actions import _fullregex

        return _fullregex(self, other)

    def isnull_action(self, other):
        from jsonutils.functions.actions import _isnull

        return _isnull(self, other)

    def length_action(self, other):
        from jsonutils.functions.actions import _length

        return _length(self, other)

    def type_action(self, other):
        from jsonutils.functions.actions import _type

        return _type(self, other)

    def key_action(self, other):
        from jsonutils.functions.actions import _key

        return _key(self, other)

    def index_action(self, other):
        from jsonutils.functions.actions import _index

        return _index(self, other)

    def path_action(self, other):
        from jsonutils.functions.actions import _path

        return _path(self, other)

    def notpath_action(self, other):
        from jsonutils.functions.actions import _notpath

        return _notpath(self, other)

    def startswith_action(self, other):
        from jsonutils.functions.actions import _startswith

        return _startswith(self, other)

    def endswith_action(self, other):
        from jsonutils.functions.actions import _endswith

        return _endswith(self, other)

    def apply_action(self, other):
        from jsonutils.functions.actions import _apply

        return _apply(self, other)

    def bool_action(self, other):
        from jsonutils.functions.actions import _bool

        return _bool(self, other)

    def nchild_action(self, other):
        from jsonutils.functions.actions import _nchild

        return _nchild(self, other)


class JSONCompose(JSONNode):
    """
    This is the base class for JSON composed objects.
    Composed objects can be dict or list instances.
    Composed objects can send queries to children (which can be also compose or singleton objects)
    """

    is_composed = True

    def __init__(self, *args, **kwargs):
        """
        By initializing instance, it assign types to child items
        """
        super().__init__(*args, **kwargs)
        self._child_objects = UUIDdict()
        self._assign_children()

    @property
    def _data(self):
        return super().json_decode

    def _assign_children(self):
        """Any JSON object can be a child for a given compose object"""
        if isinstance(self, JSONDict):
            for key, value in self.items():
                self.__setitem__(key, value)

        elif isinstance(self, JSONList):
            for index, item in enumerate(self):
                self.__setitem__(index, item)

    def path_exists(self, iterable):
        """
        Returns a boolean especifing if selected path exist in the composed object.
        """
        from jsonutils.functions.seekers import _eval_object

        if isinstance(iterable, (str, int)):
            iterable = (iterable,)
        elif isinstance(iterable, JSONPath):
            iterable = iterable.keys

        try:
            _eval_object(self, iterable)
        except (IndexError, KeyError):
            return False
        return True

    def set_path(self, path, value):
        """
        Set composed object's value on iterable path
        Example
        -------

        data = JSONObject(
            {
                "data": [
                    {
                        "A": True
                    }
                ]
            }
        )

        >> data.set_path(
            ("data",0,"B"),
            False
        )

        >> data
            {
                "data": [
                    {
                        "A": True,
                        "B": False
                    }
                ]
            }
        """

        if isinstance(path, JSONPath):
            path = path.keys

        return _set_object(self, path, value)

    def query(
        self,
        recursive_=None,
        include_parent_=None,
        stop_at_match_=None,
        native_types_=None,
        **q,
    ):
        if not isinstance(stop_at_match_, (int, type(None))):
            raise TypeError(
                f"Argument stop_at_match_ must be an integer or NoneType, not {type(stop_at_match_)}"
            )

        # within a particular parent node, each child in _child_objects registry must have a unique key

        # ---- DYNAMIC CONFIG ----
        if recursive_ is None:
            recursive_ = config.recursive_queries
        if include_parent_ is None:
            include_parent_ = config.include_parents
        if native_types_ is None:
            native_types_ = config.native_types
        # ------------------------
        queryset = QuerySet()
        if native_types_:
            queryset._native_types = True
        queryset._root = self  # the node which sends the query
        children = self._child_objects.values()
        for child in children:
            # if child satisfies query request, it will be appended to the queryset object
            check, obj = _parse_query(child, include_parent_, **q)
            if check:
                queryset.append(obj)
                if stop_at_match_ and queryset.count() >= stop_at_match_:
                    return queryset

            # if child is also a compose object, it will send the same query to its children recursively
            if child.is_composed and recursive_:
                queryset += child.query(
                    recursive_=recursive_,
                    include_parent_=include_parent_,
                    stop_at_match_=stop_at_match_,
                    native_types_=native_types_,
                    **q,
                )
        return queryset

    def get(
        self,
        recursive_=None,
        include_parent_=None,
        throw_exceptions_=None,
        native_types_=None,
        default_=dummy,
        **q,
    ):

        # ---- DYNAMIC CONFIG ----
        if recursive_ is None:
            recursive_ = config.recursive_queries
        if include_parent_ is None:
            include_parent_ = config.include_parents
        if throw_exceptions_ is None:
            throw_exceptions_ = config.query_exceptions
        if native_types_ is None:
            native_types_ = config.native_types
        # ------------------------
        query = self.query(
            recursive_=recursive_,
            include_parent_=include_parent_,
            stop_at_match_=2,
            native_types_=native_types_,
            **q,
        )

        if not query.exists():
            if default_ != dummy:
                return default_
            if throw_exceptions_:
                raise JSONQueryException("The query has not returned any result")
            else:
                return None if native_types_ else JSONNull(None)
        elif query.count() > 1:
            if throw_exceptions_:
                raise JSONQueryMultipleValues("More than one value returned by query")
            else:
                return query.first()
        else:
            return query.first()

    def annotate(self, **kwargs):
        """
        Annotate key:value pairs in each dict's child.
        If a key already exists, then ignore it.

        Example
        -------

        import jsonutils as js

        data = js.JSONObject(
            {
                "A": [
                    {
                        "A1": 1
                    },
                    {
                        "A2": 2
                    }
                ],
                "B": {
                    "B1": 1,
                    "B2": 2
                }
            }
        )

        >> data.annotate(C=3, D=4)
            {
                "A": [
                    {
                        "A1": 1,
                        "C": 3,
                        "D": 4
                    },
                    {
                        "A2": 2,
                        "C": 3,
                        "D": 4
                    }
                ],
                "B": {
                    "B1": 1,
                    "B2": 2,
                    "C": 3,
                    "D": 4
                },
                "C": 3,
                "D": 4
            }
        """

        if isinstance(self, JSONDict):
            _registered_keys = set()
            for key, value in kwargs.items():
                # if key from annotation is not in dict keys, register it
                if key not in self.keys():
                    child = JSONObject(value)
                    child._key = key
                    child.parent = self
                    child._is_annotation = True
                    self.__setitem__(key, child)
                    _registered_keys.add(child._id)

            for ch in self._child_objects.values_except(_registered_keys):
                if ch.is_composed:
                    ch.annotate(**kwargs)
        elif isinstance(self, JSONList):
            for ch in self._child_objects.values():
                if ch.is_composed:
                    ch.annotate(**kwargs)
        return self

    def query_key(
        self,
        pattern,
        recursive_=None,
        include_parent_=None,
        stop_at_match_=None,
        native_types_=None,
        **q,
    ):
        if not isinstance(stop_at_match_, (int, type(None))):
            raise TypeError(
                f"Argument stop_at_match_ must be an integer or NoneType, not {type(stop_at_match_)}"
            )

        # within a particular parent node, each child in _child_objects registry must have a unique key

        # ---- DYNAMIC CONFIG ----
        if recursive_ is None:
            recursive_ = config.recursive_queries
        if include_parent_ is None:
            include_parent_ = config.include_parents
        if native_types_ is None:
            native_types_ = config.native_types
        # ------------------------
        queryset = KeyQuerySet()
        if native_types_:
            queryset._native_types = True
        queryset._root = self  # the node which sends the query
        children = self._child_objects.values()
        for child in children:
            # if child satisfies query request, it will be appended to the queryset object
            check, obj = _parse_query_key(child, pattern, include_parent_, **q)
            if check:
                queryset.append(obj)
                if stop_at_match_ and queryset.count() >= stop_at_match_:
                    return queryset

            # if child is also a compose object, it will send the same query to its children recursively
            if child.is_composed and recursive_:
                queryset += child.query_key(
                    pattern,
                    recursive_=recursive_,
                    include_parent_=include_parent_,
                    stop_at_match_=stop_at_match_,
                    native_types_=native_types_,
                    **q,
                )
        return queryset

    def get_key(
        self,
        pattern,
        recursive_=None,
        include_parent_=None,
        throw_exceptions_=None,
        native_types_=None,
        default_=dummy,
        **q,
    ):
        # ---- DYNAMIC CONFIG ----
        if recursive_ is None:
            recursive_ = config.recursive_queries
        if include_parent_ is None:
            include_parent_ = config.include_parents
        if throw_exceptions_ is None:
            throw_exceptions_ = config.query_exceptions
        if native_types_ is None:
            native_types_ = config.native_types
        # ------------------------
        query = self.query_key(
            pattern,
            recursive_=recursive_,
            include_parent_=include_parent_,
            stop_at_match_=2,
            native_types_=native_types_,
            **q,
        )

        if not query.exists():
            if default_ != dummy:
                return default_
            if throw_exceptions_:
                raise JSONQueryException("The query has not returned any result")
            else:
                return None if native_types_ else JSONNull(None)
        elif query.count() > 1:
            if throw_exceptions_:
                raise JSONQueryMultipleValues("More than one value returned by query")
            else:
                return query.first()
        else:
            return query.first()

    def _remove_annotations(self, recursive=True):

        if isinstance(self, JSONDict):
            for key, value in list(self.items()):
                if "_is_annotation" in value.__odir__():
                    self.pop(key)
                if value.is_composed and recursive:
                    value._remove_annotations()

        elif isinstance(self, JSONList):
            for item in self:
                if item.is_composed and recursive:
                    item._remove_annotations()

    @return_value_on_exception(empty, (IndexError, KeyError))
    def eval_path(
        self, path, fail_silently=False, native_types_=None, value_on_exception=None
    ):
        """
        Evaluate JSONCompose object over a jsonpath.

        Arguments
        ---------
            path: nested path on which the object will be evaluated.
            fail_silently: if True, it will return empty in case of errors (missing paths).
            native_types_: if True, then the result will be a Python object whenever a right path is found,
                           instead of a JSONNode object.
            value_on_exception: the value which will be returned on errors, if fail_silently is True (by default, empty)
        """
        if isinstance(path, JSONPath):
            path = path.keys
        elif isinstance(path, str):
            if "/" in path:
                path = tuple(JSONPath._cast_to_int(i) for i in path.split("/") if i)
            else:
                path = (path,)

        if native_types_ is None:
            native_types_ = config.native_types

        if not isinstance(path, (tuple, list)):
            raise TypeError(
                f"path argument must be a JSONPath, str, tuple or list instance, not {type(path)}"
            )

        if native_types_:
            res = _eval_object(self, path)._data
        else:
            res = _eval_object(self, path)
        return res

    def traverse_json(self):
        """
        Traverse recursively over all json data.

        Returns
        -------
            QuerySet object
        """

        output_list = QuerySet()

        children = self._child_objects.values()
        for child in children:
            serialized_child = child._data
            output_list.append(
                JSONObject({"path": child.jsonpath.keys, "value": serialized_child})
            )
            if child.is_composed:
                output_list += child.traverse_json()

        return output_list

    def to_path(self):
        """
        Retrieve a dict of json leaf nodes paths

        Example
        -------

        data = JSONObject(
            {
                "A": [
                    {
                        "A1": 1
                    },
                    {
                        "A2": 2
                    }
                ],
                "B": 3
            }
        )

        >> data.to_path()
            {
                ('A', 0, 'A1'): 1,
                ('A', 1, 'A2'): 2,
                ('B',): 3
            }
        """
        output_dict = {}

        children = self._child_objects.values()
        for child in children:
            if child.is_leaf:
                serialized_child = child._data
                output_dict[child.jsonpath.keys] = serialized_child
            if child.is_composed:
                output_dict.update(child.to_path())

        return output_dict

    def check_valid_types(self):
        """Check if json object has valid types (not unknown types)"""

        queryset = self.query_key("*", type__="unknown")
        if queryset.exists():
            errors = [
                {
                    item._key: {"path": item.jsonpath.keys, "type": item._type}
                    for item in queryset
                }
            ]
            return False, errors
        else:
            return True, None

    def validate_schema(self, schema):
        """Validate against a json schema with defined types"""
        from jsonutils.functions.validator import _validate_data

        return _validate_data(self, schema)

    def save(self, path, create_path=True, ensure_ascii=False, indent=4, **kwargs):
        """
        Save the JSON Composed object to a file.
        Arguments
        ---------
            path: full path where to store the output file
            create_path: if True, then path to file will be created if doesn't exist
            ensure_ascii: if you want to handle unicode values
            indent: number of indents (default 4)
        """
        _path = Path(path).resolve()

        if create_path is True:
            _path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "w") as file:
            json.dump(
                self,
                file,
                cls=JSONObjectEncoder,
                ensure_ascii=ensure_ascii,
                indent=indent,
                **kwargs,
            )

    def merge(self, other, kind="inner_join"):
        """
        Makes a JSON merge with other JSON object
        """
        # TODO
        other = JSONObject(other)
        left_paths = set(self.to_path().keys())
        right_paths = set(other.to_path().keys())

        path_dict = {}

        kind = kind.lower().strip()

        if kind == "inner_join":
            resulting_set = set.intersection(left_paths, right_paths)
        elif kind == "outer_join":
            resulting_set = set.union(left_paths, right_paths)
        else:
            raise ValueError(
                f"Argument 'kind' must match one of the following options: 'inner_join', 'outer_join'"
            )

        for path in resulting_set:
            left_value = self.eval_path(path, fail_silently=True, native_types_=True)
            right_value = other.eval_path(path, fail_silently=True, native_types_=True)

            path_dict[path] = right_value or left_value

        return JSONObject.from_path(path_dict)


class JSONSingleton(JSONNode):
    """
    This is the base class for JSON singleton objects.
    A singleton object has no children
    Singleton object might be: JSONStr, JSONFloat, JSONInt, JSONBool, JSONNull.
    """

    is_composed = False

    def query(self, **kwargs):
        return QuerySet()

    def get(self, **kwargs):
        return


# ---- COMPOSE OBJECTS ----
class JSONDict(dict, JSONCompose):
    """A Dict object"""

    _DEFAULT = object()
    get = JSONCompose.get  # override get method
    _get = dict.get  # original get method
    values = JSONNode.values

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        JSONCompose.__init__(self, *args, **kwargs)

    def __dir__(self):
        # for autocompletion stuff
        if config.autocomplete_only_nodes:
            return list(self.keys())
        else:
            return list(self.keys()) + super().__dir__()

    def rename_keys(self, dic=None, inplace=False, **kwargs):
        if dic is not None:
            if not isinstance(dic, dict):
                raise TypeError(
                    f"Argument 'dic' must be a dict instance, not {type(dic)}"
                )
            if kwargs:
                raise ValueError(
                    "The two input arguments `dic` and `kwargs` cannot be passed at the same time"
                )
            if inplace:
                _rename_keys_inplace(self, dic)
                return
            else:
                return _rename_keys(self, dic)
        else:
            if not kwargs:
                raise ValueError(
                    "You must set one of the two arguments: `dic` or `kwargs`"
                )
            if inplace:
                _rename_keys_inplace(self, kwargs)
                return
            else:
                return _rename_keys(self, kwargs)

    def __getattr__(self, name):
        try:
            return self.__getitem__(name)
        except KeyError:  # if a key error is thrown, then it will call __dir__
            return

    def __setitem__(self, k, v):
        """
        When setting a new child, we must follow this steps:
            1 - Initialize child
            2 - Remove old child item from _child_objects dictionary
            3 - Assign new child item to _child_objects dictionary
        """

        # ---- initalize child ----
        child = JSONObject(v)
        child._key = k
        child.parent = self

        # ---- remove old child ----
        prev_child = super().get(k)  # old child (maybe None)
        try:
            self._child_objects.pop(
                prev_child._id, None
            )  # we must remove current child item. If not, it would still appear in queries
        except AttributeError:
            pass

        # ---- assign new child ----
        self._child_objects[child._id] = child

        return super().__setitem__(k, child)

    def __setattr__(self, name, value):
        """To define behaviour when setting an atributte. It must register a new node if not a reserved keyword"""

        if name in JSONObject._RESERVED_ATTRIBUTES:
            return super().__setattr__(name, value)
        else:
            return self.__setitem__(name, value)

    def copy(self):
        cls = self.__class__
        obj = cls(self.json_decode)
        return obj

    def get_fields(self, *args, inplace=False):
        """
        Returns a subset of this dict with selected fields, if it exists
        """
        # TODO add test
        if inplace:
            for key in list(self):
                if key not in args:
                    self.pop(key)
            return

        obj = {k: v for k, v in self.items() if k in args}
        return JSONDict(obj)

    def get_fields_except(self, *args, inplace=False):
        """
        Returns those fields from the dict that are not contained in *args
        """
        if inplace:
            for key in list(self):
                if key in args:
                    self.pop(key)
            return

        obj = {k: v for k, v in self.items() if k not in args}
        return JSONDict(obj)

    def pop(self, key, default=_DEFAULT):
        """
        When removing a key from dict, we also must unregister the corresponding child
        from _child_objects dictionary
        """
        if key in self or default is self._DEFAULT:
            child = self[key]  # getting the child
            del self[key]
            self._child_objects.pop(child._id, None)  # unregister child
            return child
        else:
            return default

    def to_django_model(
        self,
        model,
        manager="objects",
        kind="create",
        get_fields=(),
        default_fields=(),
        fail_silently=False,
    ):
        """
        Translates a JSON dict to Django model new instance
        Arguments
        ---------
            manager:
            kind:
            get_fields: the subset of fields which will be gathered in the get request
        """

        def _check_type(arg):
            if isinstance(arg, str):
                return (arg,)
            elif isinstance(arg, (list, tuple)):
                return arg
            elif isinstance(arg, type(None)):
                return ()
            else:
                raise TypeError(f"Argument '{arg}' must be an iterable instance")

        get_fields = _check_type(get_fields)
        default_fields = _check_type(default_fields)

        return _to_django_model(
            self,
            model,
            manager=manager,
            kind=kind,
            get_fields=get_fields,
            default_fields=default_fields,
            fail_silently=fail_silently,
        )

    # ---- COMPARISON METHODS ----
    def __eq__(self, other):
        return super().__eq__(other)

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False


class JSONList(list, JSONCompose):
    """A list object"""

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        JSONCompose.__init__(self, *args, **kwargs)

    def append(self, item, serialize_nodes=True):

        child = JSONObject(item, serialize_nodes=serialize_nodes)
        child._index = self.__len__()
        child.parent = self

        self._child_objects[child._id] = child
        return super().append(child)

    def copy(self):
        cls = self.__class__
        obj = cls(self.json_decode)
        return obj

    def length(self):
        return self.__len__()

    def __dir__(self):
        if config.autocomplete_only_nodes:
            return [f"_{i}" for i in range(len(self))]
        else:
            return [f"_{i}" for i in range(len(self))] + super().__dir__()

    def __getattr__(self, name):
        """We access child list items by _0, _1, etc"""

        try:
            return self.__getitem__(int(name.replace("_", "")))
        except ValueError:
            raise AttributeError

    def __setattr__(self, name, value):
        """To define behaviour when setting an atributte. It must register a new node if not a reserved keyword"""

        if name in JSONObject._RESERVED_ATTRIBUTES:
            return super().__setattr__(name, value)
        else:
            name = int(name.replace("_", ""))
            return self.__setitem__(name, value)

    def __setitem__(self, index, item):

        # ---- initialize child ----
        child = JSONObject(item)
        child._index = index
        child.parent = self

        # ---- remove old child ----
        prev_child = super().__getitem__(index)
        try:
            self._child_objects.pop(prev_child._id, None)
        except AttributeError:
            pass

        # ---- assign new child ----
        self._child_objects[child._id] = child

        return super().__setitem__(index, child)

    # ---- COMPARISON METHODS ----
    def __eq__(self, other):
        return super().__eq__(other)

    def __gt__(self, other):
        result = super().__gt__(other)
        if result == NotImplemented:
            return False
        else:
            return result

    def __ge__(self, other):
        result = super().__ge__(other)
        if result == NotImplemented:
            return False
        else:
            return result

    def __lt__(self, other):
        result = super().__lt__(other)
        if result == NotImplemented:
            return False
        else:
            return result

    def __le__(self, other):
        result = super().__le__(other)
        if result == NotImplemented:
            return False
        else:
            return result


# ---- SINGLETON OBJECTS ----
class JSONStr(str, JSONSingleton):
    def __new__(cls, string):
        obj = super().__new__(cls, string)
        obj._data = str(string)
        return obj

    # converters
    def to_float(self, **kwargs):
        """
        Try to parse a python float64 from self string.
        Examples:
        --------
        >> from base import JSONStr
        >> JSONStr(" $5.3USD ").to_float()
            5.3
        >> JSONStr(" -$ 4,450,326.58 ").to_float()
            -4450326.58
        """

        return parse_float(self, **kwargs)

    def to_datetime(self, **kwargs):
        """Try to parse an aware datetime object from self string"""

        return parse_datetime(self, **kwargs)

    def to_timestamp(self, **kwargs):
        """Try to parse a POSIX timestamp string from self string"""

        return parse_timestamp(self, **kwargs)

    def to_bool(self):
        """Trye to parse a bool object from self string."""

        return parse_bool(self)

    def __hash__(self):
        return super().__hash__()

    # comparison magic methods
    # if data types are not compatible, then return False (no error thrown)
    # when querying, other will correspond to target query value (ex: Float__gt=<other>)
    # TODO implement clever parsing option
    def __eq__(self, other):
        # if target_value is a number, we first convert self str instance to float
        if isinstance(other, bool):
            try:
                return self.to_bool() == other
            except Exception:
                return False
        elif isinstance(other, (float, int)):
            try:
                return self.to_float() == other
            except Exception:
                return False
        # if target_value is a datetime
        elif isinstance(other, (date, datetime)):
            try:
                return self.to_datetime() == parse_datetime(other)
            except Exception:
                return False
        # if target_value is a str
        elif isinstance(other, str):
            if parse_datetime(
                other, only_check=True
            ):  # if target value is a datetime string
                try:
                    return self.to_datetime() == parse_datetime(other)
                except Exception:
                    return False
            else:
                try:
                    return super().__eq__(other)
                except Exception:
                    return False
        # otherwise (maybe list, dict, none)
        else:
            return False

    def __gt__(self, other):
        if isinstance(other, bool):
            return False
        # if target_value is a number
        if isinstance(other, (float, int)):
            try:
                return self.to_float() > other
            except Exception:
                return False
        # if target_value is a datetime
        elif isinstance(other, datetime):
            try:
                return self.to_datetime() > parse_datetime(other)
            except Exception:
                return False
        # if target_value is a str
        elif isinstance(other, str):
            if parse_datetime(
                other, only_check=True
            ):  # if target value is a datetime string
                try:
                    return self.to_datetime() > parse_datetime(other)
                except Exception:
                    return False
            else:
                try:
                    return super().__gt__(other)
                except Exception:
                    return False
        # otherwise (maybe list, dict, none, bool)
        else:
            return False

    def __ge__(self, other):
        if isinstance(other, bool):
            return False
        # if target_value is a number
        if isinstance(other, (float, int)):
            try:
                return self.to_float() >= other
            except Exception:
                return False
        # if target_value is a datetime
        elif isinstance(other, datetime):
            try:
                return self.to_datetime() >= parse_datetime(other)
            except Exception:
                return False
        # if target_value is a str
        elif isinstance(other, str):
            if parse_datetime(
                other, only_check=True
            ):  # if target value is a datetime string
                try:
                    return self.to_datetime() >= parse_datetime(other)
                except Exception:
                    return False
            else:
                try:
                    return super().__ge__(other)
                except Exception:
                    return False
        # otherwise (maybe list, dict, none, bool)
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, bool):
            return False
        # if target_value is a number
        if isinstance(other, (float, int)):
            try:
                return self.to_float() < other
            except Exception:
                return False
        # if target_value is a datetime
        elif isinstance(other, datetime):
            try:
                return self.to_datetime() < parse_datetime(other)
            except Exception:
                return False
        # if target_value is a str
        elif isinstance(other, str):
            if parse_datetime(
                other, only_check=True
            ):  # if target value is a datetime string
                try:
                    return self.to_datetime() < parse_datetime(other)
                except Exception:
                    return False
            else:
                try:
                    return super().__lt__(other)
                except Exception:
                    return False
        # otherwise (maybe list, dict, none, bool)
        else:
            return False

    def __le__(self, other):
        if isinstance(other, bool):
            return False
        # if target_value is a number
        if isinstance(other, (float, int)):
            try:
                return self.to_float() <= other
            except Exception:
                return False
        # if target_value is a datetime
        elif isinstance(other, datetime):
            try:
                return self.to_datetime() <= parse_datetime(other)
            except Exception:
                return False
        # if target_value is a str
        elif isinstance(other, str):
            if parse_datetime(
                other, only_check=True
            ):  # if target value is a datetime string
                try:
                    return self.to_datetime() <= parse_datetime(other)
                except Exception:
                    return False
            else:
                try:
                    return super().__le__(other)
                except Exception:
                    return False
        # otherwise (maybe list, dict, none, bool)
        else:
            return False


class JSONFloat(float, JSONSingleton):
    def __new__(cls, fl):
        obj = super().__new__(cls, fl)
        obj._data = float(fl)
        return obj

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        try:
            return super().__eq__(parse_float(other))
        except Exception:
            return False

    def __gt__(self, other):
        try:
            return super().__gt__(parse_float(other))
        except Exception:
            return False

    def __ge__(self, other):
        try:
            return super().__ge__(parse_float(other))
        except Exception:
            return False

    def __lt__(self, other):
        try:
            return super().__lt__(parse_float(other))
        except Exception:
            return False

    def __le__(self, other):
        try:
            return super().__le__(parse_float(other))
        except Exception:
            return False


class JSONInt(int, JSONSingleton):
    def __new__(cls, i):
        obj = super().__new__(cls, i)
        obj._data = int(i)
        return obj

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        try:
            return super().__float__().__eq__(parse_float(other))
        except Exception:
            return False

    def __gt__(self, other):
        try:
            return super().__float__().__gt__(parse_float(other))
        except Exception:
            return False

    def __ge__(self, other):
        try:
            return super().__float__().__ge__(parse_float(other))
        except Exception:
            return False

    def __lt__(self, other):
        try:
            return super().__float__().__lt__(parse_float(other))
        except Exception:
            return False

    def __le__(self, other):
        try:
            return super().__float__().__le__(parse_float(other))
        except Exception:
            return False


class JSONBool(JSONSingleton):
    def __init__(self, data):

        if not isinstance(data, bool):
            raise TypeError(f"Argument is not a valid boolean type: {type(data)}")
        super().__init__()
        self._data = data

    def __repr__(self):
        return str(self._data)

    def __str__(self):
        return self.__repr__()

    def __bool__(self):
        return self._data

    def __eq__(self, other):

        try:
            return self._data == parse_bool(other)
        except Exception:
            return False

    def __ne__(self, other):
        try:
            return self._data != parse_bool(other)
        except Exception:
            return False

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False


class JSONNull(JSONSingleton):
    def __init__(self, data):

        if not isinstance(data, type(None)):
            raise TypeError(f"Argument is not a valid None type: {type(data)}")
        super().__init__()
        self._data = data

    def __repr__(self):
        return "None"

    def __str__(self):
        return self.__repr__()

    def __bool__(self):
        return False

    def __eq__(self, other):
        try:
            return self._data == other
        except Exception:
            return False

    def __ne__(self, other):
        try:
            return self._data != other
        except Exception:
            return False

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False


class JSONUnknown(JSONSingleton):
    """Unknown object"""

    def __init__(self, data):
        super().__init__()
        self._data = data
        self._type = type(data)



import unittest
from jsonutils.base import JSONObject
import jsonutils as js


class JsonTest(unittest.TestCase):
    def test_read_html_with_headers(self):
        table = """\
<table id="id-table">
  <tr>
    <th>Company</th>
    <th>Contact</th>
    <th>Country</th>
  </tr>
  <tr>
    <td>Alfreds Futterkiste</td>
    <td>Maria Anders</td>
    <td>Germany</td>
  </tr>
  <tr>
    <td>Centro comercial Moctezuma</td>
    <td>Francisco Chang</td>
    <td>Mexico</td>
  </tr>
</table>\
"""

        self.assertEqual(
            JSONObject.read_html_table(table, attrs={"id": "id-table"}),
            JSONObject(
                [
                    {
                        "Company": "Alfreds Futterkiste",
                        "Contact": "Maria Anders",
                        "Country": "Germany",
                    },
                    {
                        "Company": "Centro comercial Moctezuma",
                        "Contact": "Francisco Chang",
                        "Country": "Mexico",
                    },
                ]
            ),
        )

    def test_html_without_headers(self):
        table = """\
<table id="id-table">
  <tr>
    <td>Alfreds Futterkiste</td>
    <td>Maria Anders</td>
    <td>Germany</td>
  </tr>
  <tr>
    <td>Centro comercial Moctezuma</td>
    <td>Francisco Chang</td>
    <td>Mexico</td>
  </tr>
</table>\
"""
        self.assertEqual(
            JSONObject.read_html_table(table, attrs={"id": "id-table"}),
            JSONObject(
                [
                    {
                        "0": "Alfreds Futterkiste",
                        "1": "Maria Anders",
                        "2": "Germany",
                    },
                    {
                        "0": "Centro comercial Moctezuma",
                        "1": "Francisco Chang",
                        "2": "Mexico",
                    },
                ]
            ),
        )

    def test_read_links(self):
        table = """\
<table id="id-table">
  <tr>
    <td>Alfreds Futterkiste</td>
    <td><a href="/es/endpoint">Maria Anders</a></td>
    <td>Germany</td>
  </tr>
  <tr>
    <td>Centro comercial Moctezuma</td>
    <td>Francisco Chang</td>
    <td>Mexico</td>
  </tr>
</table>\
"""
        self.assertEqual(
            JSONObject.read_html_table(table, parse_links=True),
            JSONObject(
                [
                    {
                        "0": "Alfreds Futterkiste",
                        "1": "/es/endpoint",
                        "2": "Germany",
                    },
                    {
                        "0": "Centro comercial Moctezuma",
                        "1": "Francisco Chang",
                        "2": "Mexico",
                    },
                ]
            ),
        )
        self.assertEqual(
            JSONObject.read_html_table(
                table, parse_links=True, link_prefix="/pre/prefix/"
            ),
            JSONObject(
                [
                    {
                        "0": "Alfreds Futterkiste",
                        "1": "/pre/prefix/es/endpoint",
                        "2": "Germany",
                    },
                    {
                        "0": "Centro comercial Moctezuma",
                        "1": "Francisco Chang",
                        "2": "Mexico",
                    },
                ]
            ),
        )



import unittest
from unittest.case import skip

import jsonutils as js
from jsonutils.base import JSONObject


class SingleQueryTest(unittest.TestCase):
    def test_bool_query(self):
        test = JSONObject(
            {
                "True": 1,
                "Comp": [
                    {"True": 1, "False": 0},
                    {"True": "0", "False": ()},
                    {"True": (0,), "False": []},
                    {"True": " ", "False": {}},
                    {"True": True, "False": False},
                    {"True": [0], "False": None},
                ],
            }
        )
        self.assertListEqual(
            test.query(True__bool=True).jsonpaths(),
            [
                "True",
                "Comp/0/True",
                "Comp/1/True",
                "Comp/2/True",
                "Comp/3/True",
                "Comp/4/True",
                "Comp/5/True",
            ],
        )
        self.assertListEqual(
            test.query(False__bool=False).jsonpaths(),
            [
                "Comp/0/False",
                "Comp/1/False",
                "Comp/2/False",
                "Comp/3/False",
                "Comp/4/False",
                "Comp/5/False",
            ],
        )



from functools import wraps

import jsonutils.config as config


class dummy:
    pass


def catch_exceptions(func):
    """
    When applied on a function, prevent it from throwing exceptions, if argument 'fail_silently' is set to True (False by default)
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get("fail_silently") is True:
            try:
                return func(*args, **kwargs)
            except Exception:
                return
        else:
            return func(*args, **kwargs)

    return wrapper


def return_str_or_datetime(func):
    """
    This decorator will be applied to the parse_datetime function,
    so that it returns strings in isoformat format whenever the
    return_string argument is set to True.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get("return_string") is True:
            return func(*args, **kwargs).isoformat()
        else:
            return func(*args, **kwargs)

    return wrapper


def return_value_on_exception(value, exception):
    """
    When applied on a function, it will return `value` if selected `exception` is raised
    and argument `fail_silently` is passed as True.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if kwargs.get("fail_silently") is True:
                try:
                    return func(*args, **kwargs)
                except exception:
                    if (
                        selected_return_value := kwargs.get("value_on_exception", dummy)
                    ) is not dummy:
                        return selected_return_value
                    return value
            else:
                return func(*args, **kwargs)

        return wrapper

    return decorator


def global_config(fun=None, **config_kw):
    """
    Applies a certain global configuration to the function on which it acts.
    By default, if no arguments are entered, it will use native_types=True, query_exceptions=False.

    Example
    -------

    from jsonutils.functions.decorators import global_config
    from jsonutils import JSONObject

    data = JSONObject(
        [
            {
                "A": 1,
                "B": 2
            },
            {
                "A": 1,
                "B": 3
            },
        ]
    )

    @global_config
    def query_1(data):
        return data.get(A=1)

    >>> query_1(data)   # this will use native_types=True, query_exceptions=False
        1               # so, this is a python `int`, not a JSONInt

    @global_config(query_exceptions=True)
    def query_2(data):
        return data.get(A=1)

    >>> query_2(data)   # now this user query_exceptions=True, so it will throw a multiple values exception.
        JSONQueryMultipleValues: More than one value returned by query
    """

    if not config_kw:
        config_kw = dict(native_types=True, query_exceptions=False)

    def _set_config(dic):
        for k, v in dic.items():
            setattr(config, k, v)

    INITIAL_CONFIG = {k: getattr(config, k) for k in config_kw}

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            _set_config(config_kw)
            try:
                result = func(*args, **kwargs)
            finally:
                _set_config(INITIAL_CONFIG)
            return result

        return wrapper

    if fun is not None:
        return decorator(fun)

    return decorator


def return_native_types(func):
    """ """

    @wraps(func)
    def wrapper(self):
        res = func(self)
        if self._native_types:
            return res._data
        return res

    return wrapper
